@@ -100,5 +100,9 @@ int main(int argc, char **argv) {
      * Cleanup function for the XML library.
      */
     xmlCleanupParser();
+    /*
+     * this is to debug memory for regression tests
+     */
+    xmlMemoryDump();
     return(0);
 }
