@@ -1,14 +1,17 @@
 /*
- * parser.h : Interfaces, constants and types related to the XML parser.
+ * Summary: the core parser module
+ * Description: Interfaces, constants and types related to the XML parser
  *
- * See Copyright for the status of this software.
+ * Copy: See Copyright for the status of this software.
  *
- * daniel@veillard.com
+ * Author: Daniel Veillard
  */
 
 #ifndef __XML_PARSER_H__
 #define __XML_PARSER_H__
 
+#include <stdarg.h>
+
 #include <libxml/xmlversion.h>
 #include <libxml/tree.h>
 #include <libxml/dict.h>
@@ -16,6 +19,7 @@
 #include <libxml/valid.h>
 #include <libxml/entities.h>
 #include <libxml/xmlerror.h>
+#include <libxml/xmlstring.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -801,69 +805,6 @@ XMLPUBFUN int XMLCALL
 		xmlParserInputGrow	(xmlParserInputPtr in,
 					 int len);
 
-/*
- * xmlChar handling
- */
-XMLPUBFUN xmlChar * XMLCALL	
-		xmlStrdup		(const xmlChar *cur);
-XMLPUBFUN xmlChar * XMLCALL	
-		xmlStrndup		(const xmlChar *cur,
-					 int len);
-XMLPUBFUN xmlChar * XMLCALL	
-		xmlCharStrndup		(const char *cur,
-					 int len);
-XMLPUBFUN xmlChar * XMLCALL	
-		xmlCharStrdup		(const char *cur);
-XMLPUBFUN xmlChar * XMLCALL	
-		xmlStrsub		(const xmlChar *str,
-					 int start,
-					 int len);
-XMLPUBFUN const xmlChar * XMLCALL	
-		xmlStrchr		(const xmlChar *str,
-					 xmlChar val);
-XMLPUBFUN const xmlChar * XMLCALL	
-		xmlStrstr		(const xmlChar *str,
-					 const xmlChar *val);
-XMLPUBFUN const xmlChar * XMLCALL	
-		xmlStrcasestr		(const xmlChar *str,
-					 xmlChar *val);
-XMLPUBFUN int XMLCALL		
-		xmlStrcmp		(const xmlChar *str1,
-					 const xmlChar *str2);
-XMLPUBFUN int XMLCALL		
-		xmlStrncmp		(const xmlChar *str1,
-					 const xmlChar *str2,
-					 int len);
-XMLPUBFUN int XMLCALL		
-		xmlStrcasecmp		(const xmlChar *str1,
-					 const xmlChar *str2);
-XMLPUBFUN int XMLCALL		
-		xmlStrncasecmp		(const xmlChar *str1,
-					 const xmlChar *str2,
-					 int len);
-XMLPUBFUN int XMLCALL		
-		xmlStrEqual		(const xmlChar *str1,
-					 const xmlChar *str2);
-XMLPUBFUN int XMLCALL
-		xmlStrQEqual		(const xmlChar *pref,
-					 const xmlChar *name,
-					 const xmlChar *str);
-XMLPUBFUN int XMLCALL		
-		xmlStrlen		(const xmlChar *str);
-XMLPUBFUN xmlChar * XMLCALL	
-		xmlStrcat		(xmlChar *cur,
-					 const xmlChar *add);
-XMLPUBFUN xmlChar * XMLCALL	
-		xmlStrncat		(xmlChar *cur,
-					 const xmlChar *add,
-					 int len);
-
-XMLPUBFUN int XMLCALL	
-		xmlStrPrintf		(xmlChar *buf,
-					 int len,
-					 const xmlChar *msg,
-					 ...);
-
 /*
  * Basic parsing Interfaces
  */
@@ -1107,6 +1048,12 @@ typedef enum {
 
 XMLPUBFUN void XMLCALL
 		xmlCtxtReset		(xmlParserCtxtPtr ctxt);
+XMLPUBFUN int XMLCALL
+		xmlCtxtResetPush	(xmlParserCtxtPtr ctxt,
+					 const char *chunk,
+					 int size,
+					 const char *filename,
+					 const char *encoding);
 XMLPUBFUN int XMLCALL
 		xmlCtxtUseOptions	(xmlParserCtxtPtr ctxt,
 					 int options);
