@@ -73,6 +73,16 @@
 #define SOCKET int
 #endif
 
+
+#ifdef __MINGW32__
+#define _WINSOCKAPI_
+#include <wsockcompat.h>
+#include <winsock2.h>
+#undef SOCKLEN_T
+#define SOCKLEN_T unsigned int
+#endif
+
+
 #include <libxml/globals.h>
 #include <libxml/xmlerror.h>
 #include <libxml/xmlmemory.h>
@@ -85,10 +95,18 @@
  * A couple portability macros
  */
 #ifndef _WINSOCKAPI_
+#ifndef __BEOS__
 #define closesocket(s) close(s)
+#endif
 #define SOCKET int
 #endif
 
+#ifdef __BEOS__
+#ifndef PF_INET
+#define PF_INET AF_INET
+#endif
+#endif
+
 #ifndef SOCKLEN_T
 #define SOCKLEN_T unsigned int
 #endif
@@ -131,6 +149,8 @@ typedef struct xmlNanoHTTPCtxt {
     char *contentType;	/* the MIME type for the input */
     char *location;	/* the new URL in case of redirect */
     char *authHeader;	/* contents of {WWW,Proxy}-Authenticate header */
+    char *encoding;	/* encoding extracted from the contentType */
+    char *mimeType;	/* Mime-Type extracted from the contentType */
 } xmlNanoHTTPCtxt, *xmlNanoHTTPCtxtPtr;
 
 static int initialized = 0;
@@ -141,6 +161,18 @@ static unsigned int timeout = 60;/* the select() timeout in seconds */
 int xmlNanoHTTPFetchContent( void * ctx, char ** ptr, int * len );
 int xmlNanoHTTPContentLength( void * ctx );
 
+/**
+ * xmlHTTPErrMemory:
+ * @extra:  extra informations
+ *
+ * Handle an out of memory condition
+ */
+static void
+xmlHTTPErrMemory(const char *extra)
+{
+    __xmlSimpleError(XML_FROM_HTTP, XML_ERR_NO_MEMORY, NULL, NULL, extra);
+}
+
 /**
  * A portability function
  */
@@ -272,8 +304,8 @@ xmlNanoHTTPScanURL(xmlNanoHTTPCtxtPtr ctxt, const char *URL) {
     while (1) {
 	if ((strchr (cur, '[') && !strchr (cur, ']')) ||
 		(!strchr (cur, '[') && strchr (cur, ']'))) {
-	    xmlGenericError (xmlGenericErrorContext, "\nxmlNanoHTTPScanURL: %s",
-		    "Syntax Error\n");
+	    __xmlIOErr(XML_FROM_HTTP, XML_HTTP_URL_SYNTAX, 
+	               "Syntax Error\n");
 	    return;
 	}
 
@@ -283,7 +315,7 @@ xmlNanoHTTPScanURL(xmlNanoHTTPCtxtPtr ctxt, const char *URL) {
 		buf[indx++] = *cur++;
     
 	    if (!strchr (buf, ':')) {
-		xmlGenericError (xmlGenericErrorContext, "\nxmlNanoHTTPScanURL: %s",
+		__xmlIOErr(XML_FROM_HTTP, XML_HTTP_USE_IP,
 			"Use [IPv6]/IPv4 format\n");
 		return;
 	    }
@@ -392,8 +424,7 @@ xmlNanoHTTPScanProxy(const char *URL) {
     while (1) {
 	if ((strchr (cur, '[') && !strchr (cur, ']')) ||
 		(!strchr (cur, '[') && strchr (cur, ']'))) {
-	    xmlGenericError (xmlGenericErrorContext, "\nxmlNanoHTTPScanProxy: %s",
-		    "Syntax error\n");
+	    __xmlIOErr(XML_FROM_HTTP, XML_HTTP_URL_SYNTAX, "Syntax Error\n");
 	    return;
 	}
 
@@ -403,7 +434,7 @@ xmlNanoHTTPScanProxy(const char *URL) {
 		buf[indx++] = *cur++;
 
 	    if (!strchr (buf, ':')) {
-		xmlGenericError (xmlGenericErrorContext, "\nxmlNanoHTTPScanProxy: %s",
+		__xmlIOErr(XML_FROM_HTTP, XML_HTTP_USE_IP,
 			"Use [IPv6]/IPv4 format\n");
 		return;
 	    }
@@ -467,7 +498,10 @@ xmlNanoHTTPNewCtxt(const char *URL) {
     xmlNanoHTTPCtxtPtr ret;
 
     ret = (xmlNanoHTTPCtxtPtr) xmlMalloc(sizeof(xmlNanoHTTPCtxt));
-    if (ret == NULL) return(NULL);
+    if (ret == NULL) {
+        xmlHTTPErrMemory("allocating context");
+        return(NULL);
+    }
 
     memset(ret, 0, sizeof(xmlNanoHTTPCtxt));
     ret->port = 80;
@@ -496,6 +530,8 @@ xmlNanoHTTPFreeCtxt(xmlNanoHTTPCtxtPtr ctxt) {
     if (ctxt->out != NULL) xmlFree(ctxt->out);
     if (ctxt->in != NULL) xmlFree(ctxt->in);
     if (ctxt->contentType != NULL) xmlFree(ctxt->contentType);
+    if (ctxt->encoding != NULL) xmlFree(ctxt->encoding);
+    if (ctxt->mimeType != NULL) xmlFree(ctxt->mimeType);
     if (ctxt->location != NULL) xmlFree(ctxt->location);
     if (ctxt->authHeader != NULL) xmlFree(ctxt->authHeader);
     ctxt->state = XML_NANO_HTTP_NONE;
@@ -528,10 +564,7 @@ xmlNanoHTTPSend(xmlNanoHTTPCtxtPtr ctxt, const char * xmt_ptr, int outlen) {
 	    	      ( socket_errno( ) != EAGAIN ) &&
 #endif
 		        ( socket_errno( ) != EWOULDBLOCK ) ) {
-	        xmlGenericError( xmlGenericErrorContext,
-				"xmlNanoHTTPSend error:  %s",
-				strerror( socket_errno( ) ) );
-
+		__xmlIOErr(XML_FROM_HTTP, 0, "send failed\n");
 		if ( total_sent == 0 )
 		    total_sent = -1;
 		break;
@@ -579,9 +612,8 @@ xmlNanoHTTPRecv(xmlNanoHTTPCtxtPtr ctxt) {
 	if (ctxt->in == NULL) {
 	    ctxt->in = (char *) xmlMallocAtomic(65000 * sizeof(char));
 	    if (ctxt->in == NULL) {
+		xmlHTTPErrMemory("allocating input");
 	        ctxt->last = -1;
-		xmlGenericError( xmlGenericErrorContext, 
-			"xmlNanoHTTPRecv:  Error allocating input memory." );
 		return(-1);
 	    }
 	    ctxt->inlen = 65000;
@@ -605,10 +637,7 @@ xmlNanoHTTPRecv(xmlNanoHTTPCtxtPtr ctxt) {
 	    ctxt->inlen *= 2;
             ctxt->in = (char *) xmlRealloc(tmp_ptr, ctxt->inlen);
 	    if (ctxt->in == NULL) {
-	        xmlGenericError( xmlGenericErrorContext,
-				"xmlNanoHTTPRecv:  %s %d bytes.",
-				"Failed to realloc input buffer to",
-				ctxt->inlen );
+		xmlHTTPErrMemory("allocating input buffer");
 		xmlFree( tmp_ptr );
 	        ctxt->last = -1;
 		return(-1);
@@ -639,9 +668,7 @@ xmlNanoHTTPRecv(xmlNanoHTTPCtxtPtr ctxt) {
 		    return ( 0 );
 
 		default:
-		    xmlGenericError( xmlGenericErrorContext,
-		    		"xmlNanoHTTPRecv:  recv( ) failure - %s",
-				strerror( socket_errno( ) ) );
+		    __xmlIOErr(XML_FROM_HTTP, 0, "recv failed\n");
 		    return(-1);
 	    }
 	}
@@ -712,7 +739,7 @@ xmlNanoHTTPReadLine(xmlNanoHTTPCtxtPtr ctxt) {
  * Try to extract useful informations from the server answer.
  * We currently parse and process:
  *  - The HTTP revision/ return code
- *  - The Content-Type
+ *  - The Content-Type, Mime-Type and charset used
  *  - The Location for redirect processing.
  *
  * Returns -1 in case of failure, the file descriptor number otherwise
@@ -756,16 +783,56 @@ xmlNanoHTTPScanAnswer(xmlNanoHTTPCtxtPtr ctxt, const char *line) {
 	if ((*cur != 0) && (*cur != ' ') && (*cur != '\t')) return;
 	ctxt->returnValue = ret;
     } else if (!xmlStrncasecmp(BAD_CAST line, BAD_CAST"Content-Type:", 13)) {
+        const xmlChar *charset, *last, *mime;
         cur += 13;
 	while ((*cur == ' ') || (*cur == '\t')) cur++;
 	if (ctxt->contentType != NULL)
 	    xmlFree(ctxt->contentType);
 	ctxt->contentType = xmlMemStrdup(cur);
+	mime = (const xmlChar *) cur;
+	last = mime;
+	while ((*last != 0) && (*last != ' ') && (*last != '\t') &&
+	       (*last != ';') && (*last != ','))
+	    last++;
+	if (ctxt->mimeType != NULL)
+	    xmlFree(ctxt->mimeType);
+	ctxt->mimeType = (char *) xmlStrndup(mime, last - mime);
+	charset = xmlStrstr(BAD_CAST ctxt->contentType, BAD_CAST "charset=");
+	if (charset != NULL) {
+	    charset += 8;
+	    last = charset;
+	    while ((*last != 0) && (*last != ' ') && (*last != '\t') &&
+	           (*last != ';') && (*last != ','))
+		last++;
+	    if (ctxt->encoding != NULL)
+	        xmlFree(ctxt->encoding);
+	    ctxt->encoding = (char *) xmlStrndup(charset, last - charset);
+	}
     } else if (!xmlStrncasecmp(BAD_CAST line, BAD_CAST"ContentType:", 12)) {
+        const xmlChar *charset, *last, *mime;
         cur += 12;
 	if (ctxt->contentType != NULL) return;
 	while ((*cur == ' ') || (*cur == '\t')) cur++;
 	ctxt->contentType = xmlMemStrdup(cur);
+	mime = (const xmlChar *) cur;
+	last = mime;
+	while ((*last != 0) && (*last != ' ') && (*last != '\t') &&
+	       (*last != ';') && (*last != ','))
+	    last++;
+	if (ctxt->mimeType != NULL)
+	    xmlFree(ctxt->mimeType);
+	ctxt->mimeType = (char *) xmlStrndup(mime, last - mime);
+	charset = xmlStrstr(BAD_CAST ctxt->contentType, BAD_CAST "charset=");
+	if (charset != NULL) {
+	    charset += 8;
+	    last = charset;
+	    while ((*last != 0) && (*last != ' ') && (*last != '\t') &&
+	           (*last != ';') && (*last != ','))
+		last++;
+	    if (ctxt->encoding != NULL)
+	        xmlFree(ctxt->encoding);
+	    ctxt->encoding = (char *) xmlStrndup(charset, last - charset);
+	}
     } else if (!xmlStrncasecmp(BAD_CAST line, BAD_CAST"Location:", 9)) {
         cur += 9;
 	while ((*cur == ' ') || (*cur == '\t')) cur++;
@@ -825,10 +892,7 @@ xmlNanoHTTPConnectAttempt(struct sockaddr *addr)
 #ifdef DEBUG_HTTP
 	perror("socket");
 #endif
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPConnectAttempt: %s - %s",
-			"socket creation failure",
-			strerror( socket_errno( ) ) );
+	__xmlIOErr(XML_FROM_HTTP, 0, "socket failed\n");
 	return(-1);
     }
     
@@ -845,6 +909,12 @@ xmlNanoHTTPConnectAttempt(struct sockaddr *addr)
 	status = ioctl(s, FIONBIO, &enable);
     }
 #else /* VMS */
+#if defined(__BEOS__)
+	{
+		bool noblock = true;
+		status = setsockopt(s, SOL_SOCKET, SO_NONBLOCK, &noblock, sizeof(noblock));
+	}
+#else /* __BEOS__ */
     if ((status = fcntl(s, F_GETFL, 0)) != -1) {
 #ifdef O_NONBLOCK
 	status |= O_NONBLOCK;
@@ -859,13 +929,11 @@ xmlNanoHTTPConnectAttempt(struct sockaddr *addr)
 #ifdef DEBUG_HTTP
 	perror("nonblocking");
 #endif
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPConnectAttempt:  %s - %s",
-			"error setting non-blocking IO",
-			strerror( socket_errno( ) ) );
+	__xmlIOErr(XML_FROM_HTTP, 0, "error setting non-blocking IO\n");
 	closesocket(s);
 	return(-1);
     }
+#endif /* !__BEOS__ */
 #endif /* !VMS */
 #endif /* !_WINSOCKAPI_ */
 
@@ -875,10 +943,7 @@ xmlNanoHTTPConnectAttempt(struct sockaddr *addr)
 	    case EWOULDBLOCK:
 		break;
 	    default:
-		xmlGenericError( xmlGenericErrorContext,
-				"xmlNanoHTTPConnectAttempt:  %s - %s",
-				"error connecting to HTTP server",
-				strerror( socket_errno( ) ) );
+		__xmlIOErr(XML_FROM_HTTP, 0, "error connecting to HTTP server");
 		closesocket(s);
 		return(-1);
 	}
@@ -894,17 +959,12 @@ xmlNanoHTTPConnectAttempt(struct sockaddr *addr)
     {
 	case 0:
 	    /* Time out */
-	    xmlGenericError( xmlGenericErrorContext, 
-	    			"xmlNanoHTTPConnectAttempt: %s",
-				"Connect attempt timed out." );
+	    __xmlIOErr(XML_FROM_HTTP, 0, "Connect attempt timed out");
 	    closesocket(s);
 	    return(-1);
 	case -1:
 	    /* Ermm.. ?? */
-	    xmlGenericError( xmlGenericErrorContext,
-	    			"xmlNanoHTTPConnectAttempt: %s - %s",
-				"Error connecting to host",
-				strerror( socket_errno( ) ) );
+	    __xmlIOErr(XML_FROM_HTTP, 0, "Connect failed");
 	    closesocket(s);
 	    return(-1);
     }
@@ -912,28 +972,22 @@ xmlNanoHTTPConnectAttempt(struct sockaddr *addr)
     if ( FD_ISSET(s, &wfd) ) {
 	SOCKLEN_T len;
 	len = sizeof(status);
+#ifdef SO_ERROR
 	if (getsockopt(s, SOL_SOCKET, SO_ERROR, (char*)&status, &len) < 0 ) {
 	    /* Solaris error code */
-	    xmlGenericError( xmlGenericErrorContext,
-	    			"xmlNanoHTTPConnectAttempt: %s - %s",
-				"Error retrieving pending socket errors",
-				strerror( socket_errno( ) ) );
+	    __xmlIOErr(XML_FROM_HTTP, 0, "getsockopt failed\n");
 	    return (-1);
 	}
+#endif
 	if ( status ) {
+	    __xmlIOErr(XML_FROM_HTTP, 0, "Error connecting to remote host");
 	    closesocket(s);
 	    errno = status;
-	    xmlGenericError( xmlGenericErrorContext,
-	    			"xmlNanoHTTPConnectAttempt: %s - %s",
-				"Error connecting to remote host",
-				strerror( status ) );
 	    return (-1);
 	}
     } else {
 	/* pbm */
-	xmlGenericError( xmlGenericErrorContext,
-		"xmlNanoHTTPConnectAttempt:  %s\n",
-		"Select returned, but descriptor not set for connection.\n" );
+	__xmlIOErr(XML_FROM_HTTP, 0, "select failed\n");
 	closesocket(s);
 	return (-1);
     }
@@ -988,10 +1042,7 @@ xmlNanoHTTPConnectHost(const char *host, int port)
 
 	status = getaddrinfo (host, NULL, &hints, &result);
 	if (status) {
-	    xmlGenericError (xmlGenericErrorContext,
-		    "xmlNanoHTTPConnectHost:  %s '%s' - %s",
-		    "Failed to resolve host", host, gai_strerror (status));
-
+	    __xmlIOErr(XML_FROM_HTTP, 0, "getaddrinfo failed\n");
 	    return (-1);
 	}
 
@@ -1057,13 +1108,9 @@ xmlNanoHTTPConnectHost(const char *host, int port)
 		    h_err_txt = "No error text defined.";
 		    break;
 	    }
-	    xmlGenericError (xmlGenericErrorContext,
-		"xmlNanoHTTPConnectHost:  %s '%s' - %s",
-		"Failed to resolve host", host, h_err_txt);
+	    __xmlIOErr(XML_FROM_HTTP, 0, h_err_txt);
 #else
-	    xmlGenericError (xmlGenericErrorContext,
-		"xmlNanoHTTPConnectHost:  %s '%s'",
-		"Failed to resolve host", host);
+	    __xmlIOErr(XML_FROM_HTTP, 0, "Failed to resolve host");
 #endif
 	    return (-1);
 	}
@@ -1229,30 +1276,22 @@ retry:
 	ctxt = xmlNanoHTTPNewCtxt(URL);
     else {
 	ctxt = xmlNanoHTTPNewCtxt(redirURL);
+	ctxt->location = xmlMemStrdup(redirURL);
     }
 
     if ( ctxt == NULL ) {
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPMethodRedir:  %s %s.",
-			"Unable to allocate HTTP context to URI",
-			( ( redirURL == NULL ) ? URL : redirURL ) );
 	return ( NULL );
     }
 
     if ((ctxt->protocol == NULL) || (strcmp(ctxt->protocol, "http"))) {
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPMethodRedir:  %s - %s.",
-			"Not a valid HTTP URI",
-			( ( redirURL == NULL ) ? URL : redirURL ) );
+	__xmlIOErr(XML_FROM_HTTP, XML_HTTP_URL_SYNTAX, "Not a valid HTTP URI");
         xmlNanoHTTPFreeCtxt(ctxt);
 	if (redirURL != NULL) xmlFree(redirURL);
         return(NULL);
     }
     if (ctxt->hostname == NULL) {
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPMethodRedir:  %s - %s",
-			"Failed to identify host in URI",
-			( ( redirURL == NULL ) ? URL : redirURL ) );
+	__xmlIOErr(XML_FROM_HTTP, XML_HTTP_UNKNOWN_HOST,
+	           "Failed to identify host in URI");
         xmlNanoHTTPFreeCtxt(ctxt);
 	if (redirURL != NULL) xmlFree(redirURL);
         return(NULL);
@@ -1282,12 +1321,10 @@ retry:
     if (contentType && *contentType)
 	blen += strlen(*contentType) + 16;
     blen += strlen(method) + strlen(ctxt->path) + 24;
-    bp = xmlMallocAtomic(blen);
+    bp = (char*)xmlMallocAtomic(blen);
     if ( bp == NULL ) {
         xmlNanoHTTPFreeCtxt( ctxt );
-	xmlGenericError( xmlGenericErrorContext,
-			"xmlNanoHTTPMethodRedir:  %s",
-			"Error allocating HTTP header buffer." );
+	xmlHTTPErrMemory("allocating header buffer");
 	return ( NULL );
     }
 
@@ -1500,6 +1537,7 @@ xmlNanoHTTPFetch(const char *URL, const char *filename, char **contentType) {
     return(0);
 }
 
+#ifdef LIBXML_OUTPUT_ENABLED
 /**
  * xmlNanoHTTPSave:
  * @ctxt:  the HTTP context
@@ -1536,6 +1574,7 @@ xmlNanoHTTPSave(void *ctxt, const char *filename) {
     xmlNanoHTTPClose(ctxt);
     return(0);
 }
+#endif /* LIBXML_OUTPUT_ENABLED */
 
 /**
  * xmlNanoHTTPReturnCode:
@@ -1584,11 +1623,56 @@ xmlNanoHTTPAuthHeader(void *ctx) {
  */
 int
 xmlNanoHTTPContentLength( void * ctx ) {
-    xmlNanoHTTPCtxtPtr	ctxt = ctx;
+    xmlNanoHTTPCtxtPtr	ctxt = (xmlNanoHTTPCtxtPtr)ctx;
 
     return ( ( ctxt == NULL ) ? -1 : ctxt->ContentLength );
 }
 
+/**
+ * xmlNanoHTTPRedir:
+ * @ctx:  the HTTP context
+ *
+ * Provides the specified redirection URL if available from the HTTP header.
+ *
+ * Return the specified redirection URL or NULL if not redirected.
+ */
+const char *
+xmlNanoHTTPRedir( void * ctx ) {
+    xmlNanoHTTPCtxtPtr	ctxt = (xmlNanoHTTPCtxtPtr)ctx;
+
+    return ( ( ctxt == NULL ) ? NULL : ctxt->location );
+}
+
+/**
+ * xmlNanoHTTPEncoding:
+ * @ctx:  the HTTP context
+ *
+ * Provides the specified encoding if specified in the HTTP headers.
+ *
+ * Return the specified encoding or NULL if not available
+ */
+const char *
+xmlNanoHTTPEncoding( void * ctx ) {
+    xmlNanoHTTPCtxtPtr	ctxt = (xmlNanoHTTPCtxtPtr)ctx;
+
+    return ( ( ctxt == NULL ) ? NULL : ctxt->encoding );
+}
+
+/**
+ * xmlNanoHTTPMimeType:
+ * @ctx:  the HTTP context
+ *
+ * Provides the specified Mime-Type if specified in the HTTP headers.
+ *
+ * Return the specified Mime-Type or NULL if not available
+ */
+const char *
+xmlNanoHTTPMimeType( void * ctx ) {
+    xmlNanoHTTPCtxtPtr	ctxt = (xmlNanoHTTPCtxtPtr)ctx;
+
+    return ( ( ctxt == NULL ) ? NULL : ctxt->mimeType );
+}
+
 /**
  * xmlNanoHTTPFetchContent:
  * @ctx:  the HTTP context
@@ -1603,7 +1687,7 @@ xmlNanoHTTPContentLength( void * ctx ) {
  */
 int
 xmlNanoHTTPFetchContent( void * ctx, char ** ptr, int * len ) {
-    xmlNanoHTTPCtxtPtr	ctxt = ctx;
+    xmlNanoHTTPCtxtPtr	ctxt = (xmlNanoHTTPCtxtPtr)ctx;
 
     int			rc = 0;
     int			cur_lgth;
