@@ -38,7 +38,7 @@
 
 /************************************************************************
  *									*
- *			XInclude contexts handling			*
+ *			XInclude context handling			*
  *									*
  ************************************************************************/
 
@@ -50,7 +50,7 @@ typedef xmlChar *xmlURL;
 typedef struct _xmlXIncludeRef xmlXIncludeRef;
 typedef xmlXIncludeRef *xmlXIncludeRefPtr;
 struct _xmlXIncludeRef {
-    xmlChar              *URI; /* the rully resolved resource URL */
+    xmlChar              *URI; /* the fully resolved resource URL */
     xmlChar         *fragment; /* the fragment in the URI */
     xmlDocPtr		  doc; /* the parsed document */
     xmlNodePtr            ref; /* the node making the reference in the source */
@@ -70,12 +70,12 @@ struct _xmlXIncludeCtxt {
     int                 txtNr; /* number of unparsed documents */
     int                txtMax; /* size of unparsed documents tab */
     xmlNodePtr        *txtTab; /* array of unparsed text nodes */
-    xmlURL         *txturlTab; /* array of unparsed txtuments URLs */
+    xmlURL         *txturlTab; /* array of unparsed text URLs */
 
     xmlChar *             url; /* the current URL processed */
-    int                 urlNr; /* number of url stacked */
-    int                urlMax; /* size of url stack */
-    xmlChar *         *urlTab; /* url stack */
+    int                 urlNr; /* number of URLs stacked */
+    int                urlMax; /* size of URL stack */
+    xmlChar *         *urlTab; /* URL stack */
 
     int              nbErrors; /* the number of errors detected */
     int                legacy; /* using XINCLUDE_OLD_NS */
@@ -94,7 +94,7 @@ xmlXIncludeDoProcess(xmlXIncludeCtxtPtr ctxt, xmlDocPtr doc, xmlNodePtr tree);
 
 /**
  * xmlXIncludeErrMemory:
- * @extra:  extra informations
+ * @extra:  extra information
  *
  * Handle an out of memory condition
  */
@@ -115,7 +115,7 @@ xmlXIncludeErrMemory(xmlXIncludeCtxtPtr ctxt, xmlNodePtr node,
  * @ctxt: the XInclude context
  * @node: the context node
  * @msg:  the error message
- * @extra:  extra informations
+ * @extra:  extra information
  *
  * Handle an XInclude error
  */
@@ -136,7 +136,7 @@ xmlXIncludeErr(xmlXIncludeCtxtPtr ctxt, xmlNodePtr node, int error,
  * @ctxt: the XInclude context
  * @node: the context node
  * @msg:  the error message
- * @extra:  extra informations
+ * @extra:  extra information
  *
  * Emit an XInclude warning.
  */
@@ -341,7 +341,7 @@ xmlXIncludeURLPush(xmlXIncludeCtxtPtr ctxt,
  * xmlXIncludeURLPop:
  * @ctxt: the parser context
  *
- * Pops the top url from the url stack
+ * Pops the top URL from the URL stack
  */
 static void
 xmlXIncludeURLPop(xmlXIncludeCtxtPtr ctxt)
@@ -402,7 +402,7 @@ xmlXIncludeFreeContext(xmlXIncludeCtxtPtr ctxt) {
  * @ctxt:  the XInclude context
  * @URL:  the URL or file path
  * 
- * parse an document for XInclude
+ * parse a document for XInclude
  */
 static xmlDocPtr
 xmlXIncludeParseFile(xmlXIncludeCtxtPtr ctxt, const char *URL) {
@@ -419,7 +419,7 @@ xmlXIncludeParseFile(xmlXIncludeCtxtPtr ctxt, const char *URL) {
 	return(NULL);
     }
     /*
-     * try to ensure that the new document included are actually
+     * try to ensure that new documents included are actually
      * built with the same dictionary as the including document.
      */
     if ((ctxt->doc != NULL) && (ctxt->doc->dict != NULL) &&
@@ -448,15 +448,20 @@ xmlXIncludeParseFile(xmlXIncludeCtxtPtr ctxt, const char *URL) {
 
     xmlParseDocument(pctxt);
 
-    if (pctxt->wellFormed)
+    if (pctxt->wellFormed) {
         ret = pctxt->myDoc;
+	xmlDictReference(pctxt->dict);
+    }
     else {
         ret = NULL;
-        xmlFreeDoc(pctxt->myDoc);
+	if (pctxt->myDoc != NULL) {
+	    if ((ctxt->doc != NULL) && (ctxt->doc->dict != NULL) &&
+	        (pctxt->myDoc->dict == ctxt->doc->dict))
+		xmlDictReference(ctxt->doc->dict);
+	    xmlFreeDoc(pctxt->myDoc);
+	}
         pctxt->myDoc = NULL;
     }
-    if ((ctxt->doc != NULL) && (pctxt->dict == ctxt->doc->dict))
-        xmlDictReference(pctxt->dict);
     xmlFreeParserCtxt(pctxt);
     
     return(ret);
@@ -675,7 +680,7 @@ xmlXIncludeRecurseDoc(xmlXIncludeCtxtPtr ctxt, xmlDocPtr doc,
 	newctxt->urlTab = ctxt->urlTab;
 
 	/*
-	 * Inherit the documents already in use by others includes
+	 * Inherit the documents already in use by other includes
 	 */
 	newctxt->incBase = ctxt->incNr;
 	for (i = 0;i < ctxt->incNr;i++) {
@@ -818,11 +823,11 @@ xmlXIncludeCopyNodeList(xmlXIncludeCtxtPtr ctxt, xmlDocPtr target,
 }
 
 /**
- * xmlXInclueGetNthChild:
+ * xmlXIncludeGetNthChild:
  * @cur:  the node
  * @no:  the child number
  *
- * Returns the @no'th element child of @cur or NULL
+ * Returns the @n'th element child of @cur or NULL
  */
 static xmlNodePtr
 xmlXIncludeGetNthChild(xmlNodePtr cur, int no) {
@@ -844,8 +849,7 @@ xmlXIncludeGetNthChild(xmlNodePtr cur, int no) {
     return(cur);
 }
 
-xmlNodePtr xmlXPtrAdvanceNode(xmlNodePtr cur);
-
+xmlNodePtr xmlXPtrAdvanceNode(xmlNodePtr cur, int *level); /* in xpointer.c */
 /**
  * xmlXIncludeCopyRange:
  * @ctxt:  the XInclude context
@@ -856,16 +860,18 @@ xmlNodePtr xmlXPtrAdvanceNode(xmlNodePtr cur);
  * Build a node list tree copy of the XPointer result.
  *
  * Returns an xmlNodePtr list or NULL.
- *         the caller has to free the node tree.
+ *         The caller has to free the node tree.
  */
 static xmlNodePtr
 xmlXIncludeCopyRange(xmlXIncludeCtxtPtr ctxt, xmlDocPtr target,
 	                xmlDocPtr source, xmlXPathObjectPtr range) {
     /* pointers to generated nodes */
-    xmlNodePtr list = NULL, last = NULL, parent = NULL, tmp;
+    xmlNodePtr list = NULL, last = NULL, listParent = NULL;
+    xmlNodePtr tmp, tmp2;
     /* pointers to traversal nodes */
     xmlNodePtr start, cur, end;
     int index1, index2;
+    int level = 0, lastLevel = 0;
 
     if ((ctxt == NULL) || (target == NULL) || (source == NULL) ||
 	(range == NULL))
@@ -883,8 +889,37 @@ xmlXIncludeCopyRange(xmlXIncludeCtxtPtr ctxt, xmlDocPtr target,
     cur = start;
     index1 = range->index;
     index2 = range->index2;
+    /*
+     * level is depth of the current node under consideration
+     * list is the pointer to the root of the output tree
+     * listParent is a pointer to the parent of output tree (within
+       the included file) in case we need to add another level
+     * last is a pointer to the last node added to the output tree
+     * lastLevel is the depth of last (relative to the root)
+     */
     while (cur != NULL) {
-	if (cur == end) {
+	/*
+	 * Check if our output tree needs a parent
+	 */
+	if (level < 0) {
+	    while (level < 0) {
+	        tmp2 = xmlDocCopyNode(listParent, target, 0);
+	        xmlAddChild(tmp2, list);
+	        list = tmp2;
+	        listParent = listParent->parent;
+	        level++;
+	    }
+	    last = list;
+	    lastLevel = 0;
+	}
+	/*
+	 * Check whether we need to change our insertion point
+	 */
+	while (level < lastLevel) {
+	    last = last->parent;
+	    lastLevel --;
+	}
+	if (cur == end) {	/* Are we at the end of the range? */
 	    if (cur->type == XML_TEXT_NODE) {
 		const xmlChar *content = cur->content;
 		int len;
@@ -906,23 +941,25 @@ xmlXIncludeCopyRange(xmlXIncludeCtxtPtr ctxt, xmlDocPtr target,
 		if (list == NULL)
 		    return(tmp);
 		/* prune and return full set */
-		if (last != NULL)
+		if (level == lastLevel)
 		    xmlAddNextSibling(last, tmp);
 		else 
-		    xmlAddChild(parent, tmp);
+		    xmlAddChild(last, tmp);
 		return(list);
-	    } else {
+	    } else {	/* ending node not a text node */
 		tmp = xmlDocCopyNode(cur, target, 0);
-		if (list == NULL)
+		if (list == NULL) {
 		    list = tmp;
-		else {
-		    if (last != NULL)
+		    listParent = cur->parent;
+		} else {
+		    if (level == lastLevel)
 			xmlAddNextSibling(last, tmp);
-		    else
-			xmlAddChild(parent, tmp);
+		    else {
+			xmlAddChild(last, tmp);
+			lastLevel = level;
+		    }
 		}
-		last = NULL;
-		parent = tmp;
+		last = tmp;
 
 		if (index2 > 1) {
 		    end = xmlXIncludeGetNthChild(cur, index2 - 1);
@@ -939,8 +976,7 @@ xmlXIncludeCopyRange(xmlXIncludeCtxtPtr ctxt, xmlDocPtr target,
 		 */
 		continue; /* while */
 	    }
-	} else if ((cur == start) &&
-		   (list == NULL) /* looks superfluous but ... */ ) {
+	} else if (cur == start) {	/* Not at the end, are we at start? */
 	    if ((cur->type == XML_TEXT_NODE) ||
 		(cur->type == XML_CDATA_SECTION_NODE)) {
 		const xmlChar *content = cur->content;
@@ -950,27 +986,25 @@ xmlXIncludeCopyRange(xmlXIncludeCtxtPtr ctxt, xmlDocPtr target,
 		} else {
 		    if (index1 > 1) {
 			content += (index1 - 1);
+			index1 = 0;
 		    }
 		    tmp = xmlNewText(content);
 		}
 		last = list = tmp;
-	    } else {
-		if ((cur == start) && (index1 > 1)) {
-		    tmp = xmlDocCopyNode(cur, target, 0);
-		    list = tmp;
-		    parent = tmp;
-		    last = NULL;
+		listParent = cur->parent;
+	    } else {		/* Not text node */
+		tmp = xmlDocCopyNode(cur, target, 0);
+		list = last = tmp;
+		listParent = cur->parent;
+		if (index1 > 1) {	/* Do we need to position? */
 		    cur = xmlXIncludeGetNthChild(cur, index1 - 1);
+		    level = lastLevel = 1;
 		    index1 = 0;
 		    /*
 		     * Now gather the remaining nodes from cur to end
 		     */
 		    continue; /* while */
 		}
-		tmp = xmlDocCopyNode(cur, target, 1);
-		list = tmp;
-		parent = NULL;
-		last = tmp;
 	    }
 	} else {
 	    tmp = NULL;
@@ -992,28 +1026,23 @@ xmlXIncludeCopyRange(xmlXIncludeCtxtPtr ctxt, xmlDocPtr target,
 		    /* Humm, should not happen ! */
 		    break;
 		default:
-		    tmp = xmlDocCopyNode(cur, target, 1);
+		    tmp = xmlDocCopyNode(cur, target, 0);
 		    break;
 	    }
 	    if (tmp != NULL) {
-		if ((list == NULL) || ((last == NULL) && (parent == NULL)))  {
-		    return(NULL);
-		}
-		if (last != NULL)
+		if (level == lastLevel)
 		    xmlAddNextSibling(last, tmp);
 		else {
-		    xmlAddChild(parent, tmp);
-		    last = tmp;
+		    xmlAddChild(last, tmp);
+		    lastLevel = level;
 		}
+		last = tmp;
 	    }
 	}
 	/*
 	 * Skip to next node in document order
 	 */
-	if ((list == NULL) || ((last == NULL) && (parent == NULL)))  {
-	    return(NULL);
-	}
-	cur = xmlXPtrAdvanceNode(cur);
+	cur = xmlXPtrAdvanceNode(cur, &level);
     }
     return(list);
 }
@@ -1736,11 +1765,26 @@ loaded:
  */
 static int
 xmlXIncludeLoadFallback(xmlXIncludeCtxtPtr ctxt, xmlNodePtr fallback, int nr) {
+    xmlXIncludeCtxtPtr newctxt;
+    int ret = 0;
+    
     if ((fallback == NULL) || (ctxt == NULL))
 	return(-1);
+    /*
+     * It's possible that the fallback also has 'includes'
+     * (Bug 129969), so we re-process the fallback just in case
+     */
+    newctxt = xmlXIncludeNewContext(ctxt->doc);
+    if (newctxt == NULL)
+        return (-1);
+    xmlXIncludeSetFlags(newctxt, ctxt->parseFlags);
+    ret = xmlXIncludeDoProcess(newctxt, ctxt->doc, fallback->children);
+    if ((ret >=0) && (ctxt->nbErrors > 0))
+        ret = -1;
+    xmlXIncludeFreeContext(newctxt);
 
     ctxt->incTab[nr]->inc = xmlCopyNodeList(fallback->children);
-    return(0);
+    return(ret);
 }
 
 /************************************************************************
@@ -1797,10 +1841,9 @@ xmlXIncludeLoadNode(xmlXIncludeCtxtPtr ctxt, int nr) {
      */
     href = xmlXIncludeGetProp(ctxt, cur, XINCLUDE_HREF);
     if (href == NULL) {
-        /* @@@@ */
-	xmlXIncludeErr(ctxt, ctxt->incTab[nr]->ref, 
-		       XML_XINCLUDE_NO_HREF, "no href\n", NULL);
-	return(-1);
+	href = xmlStrdup(BAD_CAST ""); /* @@@@ href is now optional */
+	if (href == NULL) 
+	    return(-1);
     }
     parse = xmlXIncludeGetProp(ctxt, cur, XINCLUDE_PARSE);
     if (parse != NULL) {
@@ -1886,7 +1929,7 @@ xmlXIncludeLoadNode(xmlXIncludeCtxtPtr ctxt, int nr) {
 		((xmlStrEqual(children->ns->href, XINCLUDE_NS)) ||
 		 (xmlStrEqual(children->ns->href, XINCLUDE_OLD_NS)))) {
 		ret = xmlXIncludeLoadFallback(ctxt, children, nr);
-		if (ret == 0)
+		if (ret == 0) 
 		    break;
 	    }
 	    children = children->next;
