@@ -310,9 +310,10 @@ xmlReallocLoc(void *ptr,size_t size, const char * file, int line)
     MEMHDR *p;
     unsigned long number;
 
-    if (!xmlMemInitialized) xmlInitMemory();
     if (ptr == NULL)
-        return(NULL);
+        return(xmlMallocLoc(size, file, line));
+
+    if (!xmlMemInitialized) xmlInitMemory();
     TEST_POINT
 
     p = CLIENT_2_HDR(ptr);
@@ -769,8 +770,6 @@ xmlMemoryDump(void)
  *								*
  ****************************************************************/
 
-static int xmlInitMemoryDone = 0;
-
 /**
  * xmlInitMemory:
  *
@@ -784,9 +783,12 @@ xmlInitMemory(void)
 #ifdef HAVE_STDLIB_H
      char *breakpoint;
 #endif     
-
+    /*
+     This is really not good code (see Bug 130419).  Suggestions for
+     improvement will be welcome!
+    */
+     if (xmlMemInitialized) return(-1);
      xmlMemInitialized = 1;
-     if (xmlInitMemoryDone) return(-1);
      xmlMemMutex = xmlNewMutex();
 
 #ifdef HAVE_STDLIB_H
@@ -806,11 +808,25 @@ xmlInitMemory(void)
      xmlGenericError(xmlGenericErrorContext,
 	     "xmlInitMemory() Ok\n");
 #endif     
-     xmlInitMemoryDone = 1;
 
      return(0);
 }
 
+/**
+ * xmlCleanupMemory:
+ *
+ * Free up all the memory associated with memorys
+ */
+void
+xmlCleanupMemory(void) {
+    if (xmlMemInitialized == 0)
+        return;
+
+    xmlFreeMutex(xmlMemMutex);
+    xmlMemMutex = NULL;
+    xmlMemInitialized = 0;
+}
+
 /**
  * xmlMemSetup:
  * @freeFunc: the free() function to use
